﻿/* 
 *  TI Unresolved References 0.7
 *  Copyright (C) 2011  Ralph Schenn rschenn@t-online.de
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************
 *
 * 19.10.11 ...\( nicht als Referenz erkennen
 * 17.10.11 in GetFiles/FindReferences nicht bei jeder Iteration pruefen 
 *          ob Liste null ist
 *          Scheitern Anlage Verzeichnis in Unzip behandeln
 *          ConvertAll statt ConvertName
 * 13.10.11 Uebersetzung griechische Zeichen bei Anzeige Funktion
 * 08.10.11 Zeichenkonvertierung mit Uebergabe des zu verwendenden Dictionaries
 *          Symbole zur Anzeige Funktion konvertieren
 * 07.10.11 Konstante fuer "Zeichen nicht gefunden"
 *          Pruefung ob Anfang vor dem \ liegt in FindBegin
 *          neue Methode ByteArray2Str
 *          StringBuilder statt string
 *          bei Anzeige Inhalt der Funktion nur Parameterliste bis 
 *          EndFunc beruecksichtigen
 *          Zeilenumbrueche in Funktion darstellen
 * 06.10.11 Datei binaer lesen
 *          Kommentar als Beginn der Referenz erkennen
 *          Pruefung ob Ende rechts vom \ liegt in FindEnd
 * 05.10.11 Zeichen rechts vom \ nicht mehr betrachten
 *          gemeinsame Methode Error (MessageBox statt throw)
 *          Suche nach Beginn Funktionsaufruf bei " abbrechen
 *          nur ( stellt ein gueltiges Ende des Aufrufpfades dar
 *          Exception wenn Archiv geoeffnet ist behandeln
 *          Verzeichnisse innerhalb des Archives ignorieren
 * 30.06.11 neue Methode GetFunction und ConvertAll
 * 29.06.11 Groesse Buffer angepasst
 * 21.06.11 zip-Datei schliessen
 *          Liste in FindReferences nur dann erzeugen wenn eine gueltige
 *          Referenz gefunden wurde
 *          pruefen ob Nachbarzeichen von \ gueltig sind (andernfalls ist
 *          \ keine Referenz)
 *          Extensions Voyage 200 ergaenzt
 * 19.06.11 falsches Dateiformat behandeln
 * 06.06.11 Code <= angepasst
 *          Encoding.ASCII
 *          Beginn der Funktion muss vor \ liegen
 * 05.06.11 ohne Char.IsControl
 *          Liste mit moeglichen Funktionsbeginnen statisch
 *          relevante Extensions als List
 * 04.06.11 Array.Contains statt or-verknuepfter Bedingungen
 *          ' ' statt Char.IsWhiteSpace
 * 29.05.11 this verwenden
 *          griechische Buchstaben in Dateinamen ausschreiben
 * 25.05.11 FxCop
 * 24.05.11 Referenz kann auch in Assembler implementiert sein
 * 23.05.11 geschachtelte Funktionsaufrufe erkennen
 *          weitere Operatoren als Funktionsbeginn akzeptieren
 *          Suche Anfang und Ende in eigene Methoden ausgelagert
 * 19.05.11 Ersterstellung
 *
 */

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Forms;
using ICSharpCode.SharpZipLib.Zip;

namespace UnresRef
{
    /// <summary>
    /// Klasse zum Zugriff auf die Quelldateien mit Funktionen
    /// </summary>
    class Core : IDisposable
    {
        #region constants

        // Trennzeichen Ordner-Funktion
        public const char DivFoldFunc = '\\';

        // gesuchtes Zeichen nicht gefunden
        private const int notFound = -1;

        #endregion

        #region instance variables

        // Pfad zum TI group file
        private string path;

        // relevante Dateitypen
        private static List<string> extension = new List<string>()
        {
            ".89f", // Funktion 89
            ".89p", // Programm 89
            ".89z", // ASM 89
            ".92f", // Funktion 92
            ".92p", // Programm 92
            ".92z", // ASM 92
            ".v2f", // Funktion Voyage 200
            ".v2p", // Programm Voyage 200
            ".v2z"  // ASM Voyage 200
        };

        // Umsetzung griechische Buchstaben in Zeichenketten
        public static Dictionary<char, string> Greek = new Dictionary<char, string>()
        {
            {'\x80', "_alpha_"},
            {'\x81', "_beta_"},
            {'\x82', "_GAMMA_"},
            {'\x83', "_gamma_"},
            {'\x84', "_DELTA_"},
            {'\x85', "_delta_"},
            {'\x86', "_epsilon_"},
            {'\x87', "_zeta_"},
            {'\x88', "_theta_"},
            {'\x89', "_lambda_"},
            {'\xB5', "_mu_"},
            {'\x8A', "_xi_"},
            {'\x8D', "_rho_"},
            {'\x8F', "_sigma_"},
            {'\x90', "_tau_"},
            {'\x91', "_phi_"},
            {'\x92', "_psi_"},
            {'\x93', "_OMEGA_"},
            {'\x94', "_omega_"}
        };

        // Umsetzung griechische Buchstaben in Zeichenketten
        private static Dictionary<char, string> sym = new Dictionary<char, string>()
        {

#region Symbols

            {'\x12', "►"},          // Convert
            {'\x16', "→"},          // Store
            {'\x95', " E "},        // Exp
            {'\x96', "℮"},          // Euler
            {'\x97', " imag "},     // ??? Imaginary Unit
            {'\x98', " rad "},      // ??? Radians
            {'\x99', " trans "},    // ??? Matrix Transpose
            {'\x9C', "≤"},
            {'\x9D', "≠"},
            {'\x9E', "≥"},
            {'\x9F', " ang "},      // ??? Angle
            {'\xA8', "√"}, 
            {'\xAA', " gon "},      // ???  
            {'\xAD', "−"},          // Negative
            {'\xB4', "inv"},        // ??? Inverse Function
            {'\xBC', "dif"},        // ??? Differential
            {'\xBD', "∫"},

#endregion

#region Greek Lowercase

            {'\x80', "α"},  // Alpha
            {'\x81', "β"},  // Beta
            {'\x83', "γ"},  // Gamma
            {'\x85', "δ"},  // Delta
            {'\x86', "ε"},  // Epsilon
            {'\x87', "ζ"},  // Zeta
            {'\x88', "θ"},  // Theta
            {'\x89', "λ"},  // Lambda
            {'\xB5', "μ"},  // My
            {'\x8A', "ξ"},  // Xi
            {'\x8C', "π"},  // Pi
            {'\x8D', "ρ"},  // Rho
            {'\x8F', "σ"},  // Sigma
            {'\x90', "τ"},  // Thau
            {'\x91', "φ"},  // Phi
            {'\x94', "ω"},  // Omega

#endregion

#region Greek Uppercase

            {'\x82', "Γ"},  // Gamma
            {'\x84', "Δ"},  // Delta
            {'\x8B', "Π"},  // Pi
            {'\x8E', "Σ"},  // Sigma
            {'\x92', "Ψ"},  // Psi
            {'\x93', "Ω"}   // Omega

#endregion

            };

        // Zeichen die Funktionsbeginn sein koennen
        private static List<char> funcBegin = new List<char>() { 
            ' ',
            '\r',   // Zeilenumbruch
            '\0',   // Doppelpunkt
            '(',
            '[',    // Matrix
            '{',    // Liste
            '<',
            '\x3F', // <=
            '>',
            '\x9E', // >=
            '=',
            '\x9D', // !=
            ',',    // Trenner Parameter
            '\xAD', // negatives Vorzeichen
            '+',
            '-',
            '*',
            '/',
            '^',
            '\x9F', // Winkel
            '&',    // Stringverkettung
            '"',    // Funktionsaufruf als string (in Anfuehrungszeichen)
            '\xA9'  // Kommentar
        };

        // Zeichen die Funktionsende sein koennen
        private static List<char> funcEnd = new List<char>() { 
            '\r',   // Zeilenumbruch
            '\0',   // Doppelpunkt
            '('
        };

        #endregion

        #region public methods

        /// <summary>
        /// Pfad zum Ordner mit Funktionen speichern
        /// </summary>
        /// <param name="path">Pfad Ordner mit Funktionen</param>
        public Core(string path)
        {
            // Pfad ist gueltig
            if (Directory.Exists(path))
            {
                this.path = path;
            }
            // Pfad ist ungueltig
            else
            {
                throw new InvalidDataException("path");
            }
        }

        /// <summary>
        /// Aufraeumen
        /// </summary>
        public void Dispose()
        {
            this.path = null;
        }

        /// <summary>
        /// Dateien im Verzeichnis auflisten
        /// </summary>
        /// <returns>Liste mit Dateien</returns>
        public List<string> GetFiles()
        {
            List<string> files = new List<string>();

            // Verzeichnis mit Funktionen lesen
            DirectoryInfo di = new DirectoryInfo(this.path);

            try
            {
                // Iteration ueber alle Dateien im Verzeichnis
                foreach (FileInfo file in di.GetFiles())
                {
                    // relevante Extension gefunden
                    if (extension.Contains(Path.GetExtension(file.Name)))
                    {
                        // Datei zur Liste hinzufuegen
                        files.Add(file.Name);
                    }
                }
            }
            catch (Exception ex)
            {
                Error(ex);
            }

            // keine Datei mit relevanter Extension gefunden
            if (files.Count == 0)
            {
                files = null;
            }

            return files;
        }

        /// <summary>
        /// alle in Datei aufgerufenen Funktionen suchen
        /// </summary>
        /// <param path="path">Pfad zur Datei</param>
        /// <returns>Liste aufgerufener Funktionen</returns>
        public static List<string> GetReferences(string path)
        {
            // Pfad Datei ist gueltig
            if (File.Exists(path))
            {
                // Inhalt der Datei/Funktion lesen und 
                // Funktionsaufrufe suchen
                return FindReferences(ReadFile(path));
            }
            else
            {
                throw new InvalidDataException("path");
            }
        }

        /// <summary>
        /// Quellcode der Funktion lesen
        /// </summary>
        /// <param name="path">Pfad zur Funktion</param>
        /// <returns>Quellcode</returns>
        public static string GetFunction(string path)
        {
            // Beginn der Parameterliste
            const int begin = 88;

            string src = null;

            try
            {
                // Quellcode der Funktion lesen
                src = ReadFile(path);

                // Zeilenumbruch darstellen
                // griechische Zeichen konvertieren
                // Symbole auschreiben/konvertieren
                src = ConvertAll(sym,
                        src.Substring(begin, src.Length - begin).Replace("\r", "\r\n"));
            }
            catch
            {
                // Exception verwerfen (dann wird null geliefert)
            }

            return src;
        }

        /// <summary>
        /// Archiv entpacken
        /// </summary>
        /// <param name="archive">Pfad der zip-Datei</param>
        /// <returns>Erfolg</returns>
        public static bool Unzip(string archive)
        {
            // Verzeichnis angelegt
            bool success = false;

            // Datei im Archiv
            ZipEntry zipEntry;

            // Anzahl gelesener Bytes
            int bytesRead;

            // Name des Ordners
            string folder = Path.Combine(Path.GetDirectoryName(archive),
                Path.GetFileNameWithoutExtension(archive));

            FileStream sourceStream;

            // wenn Verzeichnis noch nicht existiert
            if (!Directory.Exists(folder))
            {
                try
                {
                    // Verzeichnis anlegen
                    Directory.CreateDirectory(folder);
                }
                catch (Exception ex)
                {
                    Error(ex);
                    return success;
                }
            }

            try
            {
                // ZIP-Datei in einen FileStream lesen
                sourceStream = new FileStream(archive, FileMode.Open);
            }
            catch (Exception ex)
            {
                Error(ex);
                return success;
            }

            // ZipInputStream für die Zip-Datei erzeugen 
            ZipInputStream zipInputStream = new ZipInputStream(sourceStream);

            try
            {
                // Alle im Archiv gespeicherten ZipEntry-Objekte durchgehen
                while ((zipEntry = zipInputStream.GetNextEntry()) != null)
                {
                    // wenn Objekt im Archiv eine Datei ist
                    if (zipEntry.IsFile)
                    {
                        // FileStream fuer die Datei erzeugen
                        FileStream resultStream =
                           File.Create(Path.Combine(folder, Path.GetFileName(zipEntry.Name)));

                        // Datei in Bloecken von maximal 1 KB in den Stream schreiben
                        byte[] buffer = new byte[1024];

                        do
                        {
                            // Den naechsten Datenblock aus dem ZipInputStream lesen
                            bytesRead = zipInputStream.Read(buffer, 0, buffer.Length);

                            // Datei konnte aus Archiv gelesen werden
                            if (bytesRead > 0)
                            {
                                // Datei in Ergebnis-Stream schreiben
                                resultStream.Write(buffer, 0, bytesRead);
                            }
                        } while (bytesRead > 0);

                        // Ergebnis-Stream schließen
                        resultStream.Close();
                    }
                }

                // zip-Datei schliessen
                sourceStream.Close();

                success = true;
            }
            // die Datei ist keine zip-Datei
            catch (ZipException zex)
            {
                // dann wird false zurueckgegeben
            }

            return success;
        }

        /// <summary>
        /// Fehlermeldung anzeigen
        /// </summary>
        /// <param name="ex">Exception</param>
        public static void Error(Exception ex)
        {
            Error(ex.Message + "\n" + ex.StackTrace);
        }

        /// <summary>
        /// Fehlermeldung anzeigen
        /// </summary>
        /// <param name="msg">Text</param>
        public static void Error(string msg)
        {
            MessageBox.Show(msg,
                "Error",
                MessageBoxButtons.OK,
                MessageBoxIcon.Error);
        }

        /// <summary>
        /// alle Symbole fuer Anzeige konvertieren
        /// </summary>
        /// <param name="dic">zu verwendendes Dictionary</param>
        /// <param name="input">Eingabe</param>
        /// <returns>Ausgabe</returns>
        public static string ConvertAll(Dictionary<char, string> dic, string input)
        {
            StringBuilder output = new StringBuilder();

            // Iteration ueber alle Zeichen
            foreach (char c in input)
            {
                output.Append(ConvertChar(dic, c));
            }

            return output.ToString();
        }

        #endregion

        #region private methods

        /// <summary>
        /// Funktion nach Referenzen durchsuchen
        /// </summary>
        /// <param name="func">Quellcode der Funktion</param>
        /// <returns>Referenzliste</returns>
        private static List<string> FindReferences(string func)
        {
            // Referenzen
            List<string> refs = new List<string>();

            // Position backslash
            int posBs = 0;

            // Anfangsindex Funktionsaufruf
            int begin = 0;

            // Endindex Funktionsaufruf
            int end = 0;

            // Datei ist nicht leer
            if (!String.IsNullOrEmpty(func))
            {
                try
                {
                    // Iteration ueber alle Zeichen der Funktion
                    foreach (char symbol in func)
                    {
                        // Trennzeichen Ordner-Funktion gefunden
                        if (symbol == DivFoldFunc &&

                            // nicht alle benachbarten Zeichen sind gueltig
                            // Zeichen vor \ pruefen (Ordnername)
                            (Char.IsLetterOrDigit(func[posBs - 1]) ||
                            Greek.ContainsKey(func[posBs - 1])) &&
                            // Backslash am Ende der Funktion gilt nicht
                            posBs < func.Length - 1)
                        {
                            // Anfang des Aufrufs suchen
                            begin = FindBegin(ref func, posBs - 1);

                            // Anfang des Aufrufs gefunden
                            if (begin != notFound)
                            {
                                // Ende des Aufrufs suchen
                                end = FindEnd(ref func, posBs + 1);

                                // Ende des Aufrufs gefunden
                                if (end != notFound)
                                {
                                    // Referenz hinzufuegen
                                    refs.Add(func.Substring(begin, end - begin));
                                }
                            }
                        }

                        posBs++;
                    }
                }
                catch (Exception ex)
                {
                    Error(ex);
                }

                // keine Referenzen gefunden
                if (refs.Count == 0)
                {
                    refs = null;
                }

                return refs;
            }
            else
            {
                throw new ArgumentNullException("func");
            }
        }

        /// <summary>
        /// Beginn des Funktionsaufrufs suchen
        /// </summary>
        /// <param name="func">Quellcode der Funktion</param>
        /// <param name="start">Index Trennzeichen Ordner-Funktion</param>
        /// <returns>Anfangsindex Funktionsaufruf, -1 bei Fehler</returns>
        private static int FindBegin(ref string func, int start)
        {
            // Beginn Funktionsaufruf
            int begin = notFound;

            // Beginn Funktionsaufruf links vom backslash suchen
            for (int i = start; i >= 0; i--)
            {
                // Beginn Funktionsaufruf gefunden
                if (funcBegin.Contains(func[i]))
                {
                    // Anfang muss links vom backslash liegen
                    if (i + 1 < start)
                    {
                        begin = i + 1;
                    }

                    break;
                }
            }

            return begin;
        }

        /// <summary>
        /// Ende des Funktionsaufrufs suchen
        /// </summary>
        /// <param name="func">Quellcode der Funktion</param>
        /// <param name="start">Index Trennzeichen Ordner-Funktion</param>
        /// <returns>Endindex Funktionsaufruf, -1 bei Fehler</returns>
        private static int FindEnd(ref string func, int start)
        {
            // Ende Funktionsaufruf
            int end = notFound;

            // Ende Funktionsaufruf rechts vom backslash suchen
            for (int i = start; i < func.Length; i++)
            {
                // Ende Funktionsaufruf gefunden
                if (funcEnd.Contains(func[i]))
                {
                    // nur ( stellt ein gueltiges Ende dar
                    // Ende muss rechts vom backslash liegen
                    if (func[i] == '(' && i > start)
                    {
                        end = i;
                    }

                    break;
                }
            }

            return end;
        }

        /// <summary>
        /// Datei binaer lesen (bei Verwendung von StreamReader wird fuer
        /// alle griechischen Buchstaben dez. 63 geliefert)
        /// </summary>
        /// <param name="path">Pfad zur Datei</param>
        /// <returns>Inhalt der Datei</returns>
        private static string ReadFile(string path)
        {
            byte[] bytes = null;

            try
            {
                // Datei mit Funktion oeffnen
                FileStream fs = new FileStream(path, FileMode.Open, FileAccess.Read);

                BinaryReader r = new BinaryReader(fs);

                // komplette Datei lesen
                bytes = r.ReadBytes((int)fs.Length);

                r.Close();
                fs.Close();
            }
            catch (Exception ex)
            {
                Error(ex);
                return null;
            }

            // Umwandlung byte[] in string
            return ByteArray2Str(bytes);
        }

        /// <summary>
        /// byte-Array in string konvertieren
        /// </summary>
        /// <param name="input">Array mit bytes</param>
        /// <returns>string</returns>
        private static string ByteArray2Str(byte[] input)
        {
            StringBuilder s = new StringBuilder();

            // Iteration ueber bytes
            foreach (byte b in input)
            {
                s.Append((char)b);
            }

            return s.ToString();
        }

        /// <summary>
        /// Zeichen konvertieren
        /// </summary>
        /// <param name="dic">Dictionary fuer Uebersetzung</param>
        /// <param name="input">Eingabe</param>
        /// <returns>Ausgabe</returns>
        private static string ConvertChar(Dictionary<char, string> dic, char input)
        {
            // zu uebersetzendes Zeichen gefunden
            if (dic.ContainsKey(input))
            {
                return dic[input];
            }
            else
            {
                return input.ToString();
            }
        }

        #endregion
    }
}
